#include <xenctrl.h>
#include <xen/xen.h>

#include <stdlib.h>
#include <errno.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/dir.h>
#include <sys/ptrace.h>
#include <sys/user.h>
#include <signal.h>
#include <sys/ioctl.h>

#include <unistd.h>
#include <string.h>
#include <fcntl.h>

#include "page_macro.h"

static enum { MODE_UNKNOWN, MODE_64, MODE_32, MODE_PAE } mode;

typedef enum { KERNEL_PGD, PROCESS_PGD } pgd_access;

/**
 * Check paging mode and Map a guest virtual address in your address space.
 * Call others map functions for x86_32 ad x86_32PAE paging mode.
 * @param[in] xc_handle Xen handle.
 * @param[in] cpu Virtual CPU id.
 * @param[in] current_domid Domain ID.
 * @param[in] guest_va Virtual guest address to map.
 * @param[in] perm Location address permissions.
 * @param[in] pgd Idenificator to indentify PGD Kernel or PGD Process access.
 * @return A virtual address in your space address that map passed guest virtual address.
		-1 if fails getContext for domU.
		NULL if fails PT walk.
*/
void *
mapmem_domu(
    int xc_handle,
    int cpu,
    int current_domid,
    void *guest_va,
    int perm,
    pgd_access pgd);

/**
 * Convert a PFN (typical Kernel address) to MFN via P2M table.
 * @param[in] xc_handle Xen handle.
 * @param[in] current_domid Domain ID.
 * @param[in] perm Location address permissions.
 * @param[in] pfn Physical Frame number to convert.
 * @return A MFN from PFN.
		-1 if fails getSgharedMemoryInfo for domU.
*/
unsigned long convert_pfn_to_mfn(
	int xc_handle,
	int current_domid,
	int perm,
	unsigned long pfn);

/**
 * Map a guest virtual address in your virtual address space. Implemented for x86_32 (Regular Paging) paging mode.
 * @param[in] xc_handle Xen handle.
 * @param[in] cpu Virtual CPU id.
 * @param[in] guest_va Virtual guest address to map.
 * @param[in] perm Location address permissions.
 * @param[in] current_domid Domain ID.
 * @param[in] pgd Idenificator to indentify PGD Kernel or PGD Process access.
 * @return A virtual address in your space address that map passed guest virtual address.
		NULL if fails PT walk or getContext for domU.
*/
static void *
mapmem_domu_rp(
	int xc_handle,
    	int cpu,
    	void *guest_va,
    	int perm, 
	int current_domid,
    	pgd_access pgd);

/**
 * Map a guest virtual address in your virtual address space. Implemented for x86_32PAE paging mode.
 * @param[in] xc_handle Xen handle.
 * @param[in] cpu Virtual CPU id.
 * @param[in] guest_va Virtual guest address to map.
 * @param[in] perm Location address permissions.
 * @param[in] current_domid Domain ID.
 * @param[in] pgd Idenificator to indentify PGD Kernel or PGD Process access.
 * @return A virtual address in your space address that map passed guest virtual address.
		NULL if fails PT walk or getContext for domU.
*/
static void *
mapmem_domu_pae(
    int xc_handle,
    int cpu,
    void *guest_va,
    int perm,
    int current_domid,
    pgd_access pgd);

/**
 * Retrieve a kernel symbol address table.
 * @param[in] symbol_t System.map path.
 * @param[in] symbol Symbol name.
 * @return A address in hexadecimal format relative to symbol passed.
 		-1 if symbol isn't found or occur an error opening System.map table.
*/
unsigned long get_kernel_address_from_symbol(
	char *symbol_t,
	char *symbol);

/**
 * Retrieve a valid MFN for CR3 Kernel register address. CR3 MFN value is written in cr3 signature parameter.
 * [NOTE] This MFN need map in your address space (xc_map_foreign_range) to access.
 * @param[in] xc_handle Xen handle.
 * @param[in] current_domid Domain ID.
 * @param[in] perm Location address permissions
 * @param[in] cr3 The cr3 address
 * @return 0 if occur no error, -1 otherwise.
*/
int get_mfn_cr3_kernel(
	int xc_handle,
	int current_domid,
	int perm,
	unsigned long *cr3);

/**
 * Retrieve a current IP for a specified domU.
 * @param[in] xc_handle Xen handle.
 * @param[in] cpu Virtual CPU id.
 * @param[in] current_domid Domain ID.
 * @return The actual IP (an address) for a domU.
		NULL if fails getContext for domU.
*/
uint32_t *
read_eip(
	int xc_handle,
	uint32_t cpu,
	uint32_t current_domid);

/**
 * Retrieve a current IP for a specified domU.
 * @param[in] xc_handle Xen handle.
 * @param[in] cpu Virtual CPU id.
 * @param[in] current_domid Domain ID.
 * @return 0 if IP is set correctly, -1 otherwise.
*/
int
write_eip(
	int xc_handle,
	uint32_t cpu,
	uint32_t current_domid,
	uint32_t new_eip);

/**
 * Verfify if an address is in Kernel space memory.
 * @param[in] address a virtual address.
 * @return 1 if is Kernel address, 0 otherwise.
*/
int is_kernel_address(
	unsigned long address);


#ifdef linux_ver2616
#define map_file "/boot/System.map-2.6.16-xenU"
#endif

#ifdef linux_ver2618
#define map_file "/boot/System.map-2.6.18-xenU"
#endif
